/* DYNAMIC LIGHTWEIGHT SC DENSE CODE. (NO linear time version) -- 
A dynamic word-based byte oriented compressor for text files based on 
(s,c)-Dense Code.

Programmed by Antonio Faria
Copyright (C) 2006  Nieves R. Brisaboa, Gonzalo Navarro, Antonio Faria and Jos R. Param
Author's contact: antonio.fari@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
   

/*-----------------------------------------------------------------------
 File      : compressor
 Function  : Main file of the compressor
 ------------------------------------------------------------------------*/


#include "compressor.h"


/*------------------------------------------------------------------
 Initilizes the structures used.
 ---------------------------------------------------------------- */
void initialize (unsigned long sizeVoc, unsigned long tamTop){
	unsigned long i;
	unsigned long N;
	unsigned long N1;
	
	N=sizeVoc;
	N1=tamTop;
	
	_memMgr = createMemoryManager(); //_memMgr.initialize();
	
	initialize_hash(N);
	positionInTH = (unsigned long*) malloc(N * sizeof(unsigned long));
	
	zeroNode = 0;
	
	//initializes top and last values for each block
	topBlock = (unsigned long*) malloc(N1 * sizeof(unsigned long));
	MAXFREQ = 1;
	topBlock[0] =zeroNode;
		
	S_value = Initial_S_value;
	C_value = B_value - S_value;

 	initializeScapeValues();
}


void initializeScapeValues() {
	
	LEN_SCAPES_C_NEW=4;
	LEN_SCAPES_C_SWAP=4;
	LEN_SCAPES_C_SCHANGE=4;
	
	CC_NEW[0]=255; CC_NEW[1]=255;  CC_NEW[2]=255;  CC_NEW[3]=S_value -3;
	CC_SWAP[0]=255; CC_SWAP[1]=255;  CC_SWAP[2]=255;  CC_SWAP[3]=S_value -2;
	CC_SCHANGE[0]=255; CC_SCHANGE[1]=255;  CC_SCHANGE[2]=255;  CC_SCHANGE[3]=S_value -1;

}


/*------------------------------------------------------------------
 * oBytes: Returns the number of bytes needed to encode the word
 * ranked "i" when using s_i value as "S"
------------------------------------------------------------------ */
unsigned long oBytes (unsigned long s_i, unsigned long i_val) {
		unsigned long k, c_i;
		unsigned long last,pot;

		//fprintf(stdFile,"\nEntering oBytes con s = %lu",s_i);
		//counts the number of bytes of the codeword to be generated
		k  	= 1;
		last  = pot = s_i;
		c_i	= B_value - s_i;

		while (last <= i_val) {
			pot *= c_i;
			last += pot;
			k++;
		}
		////fprintf(stdFile,"\n[ZERONODE=%lu]Size of i_val [%lu] when s= [%lu] is [%lu] \n",zeroNode,i_val,s_i,k);
		return k;
	}


/*------------------------------------------------------------------
 * takeAccount: Adds to TCS_{s-1}, TCS_{s}, TCS_{s+1} , the number of
 * bytes needed to encode the word ranked "j", depending on the "s_x" used
------------------------------------------------------------------ */
 void takeAccount (BufferWrite bw) {

	TCs0 +=oBytes (S_value - 1,j);
//	TCs1 +=K; //oBytes (S_value,j);
	TCs1 +=oBytes (S_value,j);

	if (TCs0  <  TCs1) {
		S_value --;  
		C_value  ++;
		encodeSValue(bw,S_value+1);
		//fprintf (stderr,"\nCHANGE IN S VALUE: GOING DOWN FROM [%lu] TO [%lu]",S_value+1,S_value);
		//fprintf (stderr,"\nzeroNode value is [%lu]",zeroNode);
		//	fprintf(stdFile,"corpus(%lu)=%lu;\n",zeroNode,S_value);
		TCs2=TCs1;
		TCs1=TCs0;
	}	
	else {
		TCs2 += oBytes (S_value +1,j);
		if (TCs2  <  TCs1 )  {
			S_value ++;
			C_value  --;
			encodeSValue(bw,S_value-1); 
			//fprintf (stderr,"\nCHANGE IN S VALUE: GOING UP FROM [%lu] TO [%lu]",S_value-1,S_value);
			//fprintf (stderr,"\nzeroNode value is [%lu]",zeroNode);
			//	fprintf(stdFile,"corpus(%lu)=%lu;\n",zeroNode,S_value);
			TCs0=TCs1;
			TCs1=TCs2;
		}
	}	
}


/*------------------------------------------------------------------
 * GetCode
 * Obtains the codeword of a word ranked in "position"
 * returns the codeword and its size by modifing
 * two global scope variables: CODEWORD and K
------------------------------------------------------------------ */

void getCode(register unsigned long position) {

	register unsigned long kk=1; //counts the number of bytes of the codeword to be generated

	if (position >= S_value) {
	   register unsigned long seguir;
		register unsigned  long i,r;
		register unsigned  long codeword;
		i=position;
		(codeword) = (i % S_value);
		seguir 	  = (i / S_value) -1;

		while (seguir >= C_value) {
			r = seguir % C_value;
			seguir = seguir/C_value -1;
			//seguir -=1;
			(codeword) = (codeword) + ( (r + S_value)<<(8*kk) ) ;
			kk ++;
		}
		(codeword) = (codeword) + ( (seguir + S_value  )<<(8*kk) ) ;
		kk ++;

		K  = kk;
		CODEWORD= codeword;
	}
	else {
		K  = kk;
		CODEWORD= position;
	}
}



void getCodeAndWrite(BufferWrite bw,register unsigned long position)  {
	char code[8];
	register unsigned long kk=1; //counts the number of bytes of the codeword to be generated
	unsigned char *codeword;

	codeword =(unsigned char *) (code+7);

	if (position >= S_value) {
	    register unsigned long seguir;
		register unsigned  long i,r;
		//register unsigned  long codeword;
		i=position;
		//(codeword) = (i % S_value);
		(*codeword--) = (i % S_value);
		seguir 	 	= (i / S_value) -1;

			while (seguir >= C_value) {
				r = seguir % C_value;
				seguir = seguir/C_value -1;
				//(codeword) = (codeword) + ( (r + S_value)<<(8*kk) ) ;
				kk ++;
				(*(codeword--)) = r + S_value;
			}
//		(codeword) = (codeword) + ( (seguir + S_value  )<<(8*kk) ) ;
		(*codeword) = (seguir + S_value  );
		kk ++;

//		K  = kk;
//		CODEWORD= codeword;

	}
	else {
//		K  = kk;
//		CODEWORD= position;
		*codeword = position;
	}

writeBW(bw,codeword,kk);
//K=kk;
}




unsigned long findCodeLeft(unsigned long codePos) {
	long x;
	//for (x=codePos;x>=0;x++) {
	x=codePos;
	while(1){
		if (hash[positionInTH[x]].codeword == codePos) return x;
		x--;
	}
	return -1; //never should occur.
}

unsigned long findCodeRight(unsigned long codePos) {
	long x;
	x=codePos;
//	for (x=codePos;x<;x++) {
	while(1) {
		if (hash[positionInTH[x]].codeword == codePos) return x;
			x++;
	}
	return -1; //never should occur.
	
}

/*------------------------------------------------------------------
 * Notifies a SValue Change to the decompressor.
------------------------------------------------------------------ */
void encodeSValue(BufferWrite bw, unsigned long oldS) {
	
	unsigned long newS;
	newS=S_value;
	S_value = oldS;
	C_value = B_value-S_value;
	
	#ifdef DEBUG_ENABLED
		SCHANGES ++;
	#endif
	//writeCodeWord(bw,&C_SCHANGE,&LEN_SCAPES_C_SCHANGE);	
	writeBW(bw,(unsigned char *)CC_SCHANGE,LEN_SCAPES_C_SCHANGE);	
	
	//getCode(B_value-newS);  //encodes c_value
	//writeCodeWord(bw,&CODEWORD,&K);	
	getCodeAndWrite (bw,B_value-newS);
	
	//performs the swap between word in pos S(or s-1) and word in pos cwPos[S] (or s-1)
	{ 
 		if (zeroNode > Initial_S_value) {		
			unsigned long j1,j2;
			unsigned long p,h;
	
	//fprintf(stderr,"\n FIND CODES ?? ENTRA");
			if (oldS > newS) {
				oldS = oldS -1;  // looks for the value s-1;
				j1=findCodeLeft(oldS);
			}
			else j1=findCodeRight(oldS);

	//fprintf(stderr,"\n FIND CODES ?? sale %ld ", j1);
						
			p=positionInTH[j1];
			
			h=positionInTH[oldS];
			j2=hash[h].codeword;
			
			swap(&(hash[p].codeword),&(hash[h].codeword));	
			
			//Notify the receiver that codewords oldS and "j2" have to be exchanged.
//			getCode(j2);  //encodes c_value
//			writeCodeWord(bw,&CODEWORD,&K);
			getCodeAndWrite (bw,j2);
		
			//fprintf(stderr,"\n send %ld, %ld, %ld", j2,CODEWORD,K);
			//fprintf(stderr,"\n intercambio %ld, %ld", hash[p].posInVocab, hash[h].posInVocab);
		}
		else {
//			getCode(S_value-1); 
//			writeCodeWord(bw,&CODEWORD,&K);
			getCodeAndWrite (bw,S_value-1);
		}
	}
	
	//end Sequence with a '\0' terminator character --> needed for searches.
	//in order a stopper can be found when horspool goes back and after a Change
	//that decreases S_value
	{unsigned char zero[1];
		zero[0]='\0';
		writeBW(bw,(unsigned char *)zero,1);
	}	
	//updates S_value
	S_value=newS;
	C_value = B_value-S_value;
	initializeScapeValues();	
}

/*------------------------------------------------------------------
 * Performs the full encoding process.
   One pass through the text:
   	- Reads a word.
   	- New word is preceded by C_new "ASCII_WORD" \0
   	- Existing word is encoding as C_code
------------------------------------------------------------------ */

void mainCompressor (char* infile, char* outfile, unsigned long N_val, unsigned long N_val2) {

	BufferRead B;
	unsigned char *text,*top, *nextbuf;
	register unsigned char *pbeg,*pend,*wordstart;
	register unsigned long size;
	BufferWrite bw;

	register unsigned long i;
	int f;

	unsigned char *aWord;
	unsigned int ONLY1BUFFER;
	register unsigned long kkk1, kkk2;
	
	
	

  	StartValid();

	// Initialization of read buffer
	f = open (infile, O_RDONLY);
	if (f == -1) {
		printf("Cannot read file %s",infile);
		exit(0);
	}

	B = bufCreate(MAX_SIZE_OF_WORD+5);    //nmero de "bytes" que hay que dejar al final del buffer siempre, por culpa

	bufSetFile (B,f);

	// Initialization of output buffer

	bw= bufCreateWriter(outfile);
	openBW(bw);

	initialize(N_val,N_val2); //initializes compressor for up to N_value words
	fprintf(stderr, "\nSTARTING COMPRESSION...");

	if (bufEof(B)) {ONLY1BUFFER=1;}else {ONLY1BUFFER=0;}


	/*** Beginning of the compression process   ***/
	if (bufEmpty(B) ) {size =0; return;}
	size = 0;

	bufCurrent (B,(char**)&text,(char**)&top);
	pbeg = text; pend = top;

       	while (( pbeg < pend ) && (zeroNode < S_value) )   {

		//parsing a word or separator
		size=0;
		wordstart = pbeg;
		if (_Valid[*pbeg]) {   //alphanumerical data

			while ( (size<MAX_SIZE_OF_WORD) && (  _Valid[*pbeg] )) {
					size++;
					pbeg++;
			}

	    }
      	else {
			if (*pbeg != ' ') { //a separator comes starting in ' ' comes, so it is a new word

				while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg])) {

					size++;
					pbeg++;
				}
			}
			else {  //a  SPACE comes, so we have to test if next character is alphanumerical or not

				pbeg++;
				if (_Valid [*pbeg] ) {
					wordstart = pbeg;
					while ( (size<MAX_SIZE_OF_WORD) && (_Valid[*pbeg] )) {
						size++;
						pbeg++;
					}

				}
				else {   // a "separator word" is parsed until an alphanumerical character is found

					size++;
					while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg] )) {
						size++;
						pbeg++;
					}
				}//else {  // a "separator word"
			}// end else  (!bufEmpty(B)) {
  		}
  		if ( pbeg >= pend ) {  //need in case of the last word been parsed further away of cend. 
  			size -= (pbeg-pend);	
  		}

		//The parsed word ...
		aWord=wordstart;
		j = search ((unsigned char *)aWord, size, &addrInTH );

			//Dynamic update of the data structures of the compressor//
			{
				register unsigned long p,f,i,h,j1;
				register unsigned long temp;

				p=addrInTH;
		
				if (j==zeroNode) {
					insertElement ((unsigned char *) aWord, size, &addrInTH);
					hash[p].len=size;

					//getCode(j);
					hash[p].codeword = j;
					
					hash[p].weight = 1;
					hash[p].posInVocab = zeroNode;

					positionInTH[zeroNode] = p;
					zeroNode++;

					writeBW(bw,(unsigned char *)CC_NEW,LEN_SCAPES_C_NEW);	
					writeBWPlusZero(bw,(unsigned char *)aWord,size);
				}
				else {
					f = hash[p].weight;
					hash[p].weight +=1;

					i = hash[p].posInVocab;									
					j1 = topBlock[f];
					h = positionInTH[j1];

					temp = positionInTH[i];
					positionInTH[i]=  positionInTH[j1];
					positionInTH[j1] = temp;

					hash[p].posInVocab =j1;
					hash[h].posInVocab =i;

					topBlock[f]=j1+1;

					//	if (hash[h].size<hash[p].size) {
					
					
					{   unsigned long tmp;
						kkk1=hash[h].codeword;
						kkk2=hash[p].codeword;
						tmp=kkk1;
						if (kkk1 > kkk2) {	
							kkk2=kkk1;
							kkk1=tmp;
						}
						
						if (  ((kkk1 < W[S_value][0]) && (kkk2 >= W[S_value][0])) ||
							  ((kkk1 < W[S_value][1]) && (kkk2 >= W[S_value][1])) ||
							  ((kkk1 < W[S_value][2]) && (kkk2 >= W[S_value][2])) 
						    )				
						kkk1=kkk2-1;
						else kkk1=kkk2;
							
							
					}
					
					
					//kkk1= oBytes(S_value,hash[h].codeword);
					//kkk2= oBytes(S_value,hash[p].codeword);
					if (kkk1 < kkk2) {				
						swap(&(hash[p].codeword),&(hash[h].codeword));

						writeBW(bw,(unsigned char *)CC_SWAP,LEN_SCAPES_C_SWAP);	
						#ifdef DEBUG_ENABLED
							SWAPS++;
						#endif

//						getCode(hash[h].codeword);
//						writeCodeWord(bw,&CODEWORD,&K);
						getCodeAndWrite (bw,hash[h].codeword);
						
					}

					//writes the codeword.
					{					
//						getCode(hash[addrInTH].codeword);
//						writeCodeWord(bw,&CODEWORD,&K);
						getCodeAndWrite (bw,hash[addrInTH].codeword);
					}

				}//end else "j <> zeronode"

			} //End of dynamic update

    	}/* inner while*/
	
	
	S_value --;  
  	C_value  ++;
  	
  	encodeSValue(bw,S_value+1);
	//fprintf (stderr,"\nCHANGE IN S VALUE: GOING DOWN FROM [%lu] TO [%lu]",S_value+1,S_value);
	//fprintf (stderr,"\nzeroNode value is [%lu]",zeroNode);
	//	fprintf(stdFile,"corpus(%lu)=%lu;\n",zeroNode,S_value);

	/************************************************************************/
	/************************************************************************/
	/*    main 2 ...                                                        */
	/************************************************************************/
	/************************************************************************/

	size = 0;

 	while (( pbeg < pend ) && ( S_value > 210))  {

		//parsing a word or separator
		size=0;
		wordstart = pbeg;
		if (_Valid[*pbeg]) {   //alphanumerical data

			while ( (size<MAX_SIZE_OF_WORD) && (  _Valid[*pbeg] )) {
					size++;
					pbeg++;
			}

	    }
      	else {
			if (*pbeg != ' ') { //a separator comes starting in ' ' comes, so it is a new word

				while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg])) {

					size++;
					pbeg++;
				}
			}
			else {  //a  SPACE comes, so we have to test if next character is alphanumerical or not

				pbeg++;
				if (_Valid [*pbeg] ) {
					wordstart = pbeg;
					while ( (size<MAX_SIZE_OF_WORD) && (_Valid[*pbeg] )) {
						size++;
						pbeg++;
					}

				}
				else {   // a "separator word" is parsed until an alphanumerical character is found

					size++;
					while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg] )) {
						size++;
						pbeg++;
					}
				}//else {  // a "separator word"
			}// end else  (!bufEmpty(B)) {
  		}


		//The parsed word ...
		aWord=wordstart;
		j = search ((unsigned char *)aWord, size, &addrInTH );

			//Dynamic update of the data structures of the compressor//
			{
				register unsigned long p,f,i,h,j1;
				register unsigned long temp;

				p=addrInTH;
		
				if (j==zeroNode) {
					insertElement ((unsigned char *) aWord, size, &addrInTH);
					hash[p].len=size;

					//getCode(j);
					hash[p].codeword = j;
					
					hash[p].weight = 1;
					hash[p].posInVocab = zeroNode;

					positionInTH[zeroNode] = p;
					zeroNode++;

					writeBW(bw,(unsigned char *)CC_NEW,LEN_SCAPES_C_NEW);	
					writeBWPlusZero(bw,(unsigned char *)aWord,size);
				}
				else {
					f = hash[p].weight;
					hash[p].weight +=1;

					i = hash[p].posInVocab;									
					j1 = topBlock[f];
					h = positionInTH[j1];

					temp = positionInTH[i];
					positionInTH[i]=  positionInTH[j1];
					positionInTH[j1] = temp;

					hash[p].posInVocab =j1;
					hash[h].posInVocab =i;

					topBlock[f]=j1+1;

					//	if (hash[h].size<hash[p].size) {
					
					{   unsigned long tmp;
						kkk1=hash[h].codeword;
						kkk2=hash[p].codeword;
						tmp=kkk1;
						if (kkk1 > kkk2) {	
							kkk2=kkk1;
							kkk1=tmp;
						}
						
						if (  ((kkk1 < W[S_value][0]) && (kkk2 >= W[S_value][0])) ||
							  ((kkk1 < W[S_value][1]) && (kkk2 >= W[S_value][1])) ||
							  ((kkk1 < W[S_value][2]) && (kkk2 >= W[S_value][2])) 
						    )				
						kkk1=kkk2-1;
						else kkk1=kkk2;
					}
					
					
					//kkk1= oBytes(S_value,hash[h].codeword);
					//kkk2= oBytes(S_value,hash[p].codeword);
					if (kkk1 < kkk2) {				
						swap(&(hash[p].codeword),&(hash[h].codeword));

						writeBW(bw,(unsigned char *)CC_SWAP,LEN_SCAPES_C_SWAP);	
						#ifdef DEBUG_ENABLED
							SWAPS++;
						#endif
//						getCode(hash[h].codeword);
//						writeCodeWord(bw,&CODEWORD,&K);
						getCodeAndWrite (bw,hash[h].codeword);
					}

					//writes the codeword.
					{					
//						getCode(hash[addrInTH].codeword);
//						writeCodeWord(bw,&CODEWORD,&K);
						getCodeAndWrite (bw,hash[addrInTH].codeword);
					}

				}//end else "j <> zeronode"

			} //End of dynamic update
			
		//getCode(j);
		TCs0 +=oBytes(S_value -1, j);
		//TCs1 +=K;  //oBytes(S_value, j);
		TCs1 +=oBytes(S_value, j);
		
		
		
		if (TCs0 < TCs1){
			S_value --;  
		  	C_value ++;
		  	TCs1 = TCs0;
		  	encodeSValue(bw,S_value+1);
			//fprintf (stderr,"\nCHANGE IN S VALUE: GOING DOWN FROM [%lu] TO [%lu]",S_value+1,S_value);
			//fprintf (stderr,"\nzeroNode value is [%lu]",zeroNode);
			
			//	fprintf(stdFile,"corpus(%lu)=%lu;\n",zeroNode,S_value);
		}

    }/* inner while*/

   	TCs2 = TCs1;




	/************************************************************************/
	/************************************************************************/
	/*    main 3 ...                                                        */
	/************************************************************************/
	/************************************************************************/


	//fprintf(stderr, "\n\n\nEntering normal process...");


	/***************************************************************************/
	/*  main 3.1, first iteration to end the pending buffer  (the first buffer)*/
	/***************************************************************************/

	while (( pbeg < pend ))  {
		//parsing a word or separator
		size=0;
		wordstart = pbeg;
		if (_Valid[*pbeg]) {   //alphanumerical data

			while ( (size<MAX_SIZE_OF_WORD) && (  _Valid[*pbeg] )) {
					size++;
					pbeg++;
			}

	    }
      	else {
			if (*pbeg != ' ') { //a separator comes starting in ' ' comes, so it is a new word

				while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg])) {

					size++;
					pbeg++;
				}
			}
			else {  //a  SPACE comes, so we have to test if next character is alphanumerical or not

				pbeg++;
				if (_Valid [*pbeg] ) {
					wordstart = pbeg;
					while ( (size<MAX_SIZE_OF_WORD) && (_Valid[*pbeg] )) {
						size++;
						pbeg++;
					}

				}
				else {   // a "separator word" is parsed until an alphanumerical character is found

					size++;
					while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg] )) {
						size++;
						pbeg++;
					}
				}//else {  // a "separator word"
			}// end else  (!bufEmpty(B)) {
  		}


		//The parsed word ...
		aWord=wordstart;
		j = search ((unsigned char *)aWord, size, &addrInTH );

			//Dynamic update of the data structures of the compressor//
			{
				register unsigned long p,f,i,h,j1;
				register unsigned long temp;

				p=addrInTH;
		
				if (j==zeroNode) {
					insertElement ((unsigned char *) aWord, size, &addrInTH);
					hash[p].len=size;

					//getCode(j);
					hash[p].codeword = j;
					
					hash[p].weight = 1;
					hash[p].posInVocab = zeroNode;

					positionInTH[zeroNode] = p;
					zeroNode++;

					writeBW(bw,(unsigned char *)CC_NEW,LEN_SCAPES_C_NEW);	
					writeBWPlusZero(bw,(unsigned char *)aWord,size);
				}
				else {
					f = hash[p].weight;
					hash[p].weight +=1;

					i = hash[p].posInVocab;									
					j1 = topBlock[f];
					h = positionInTH[j1];

					temp = positionInTH[i];
					positionInTH[i]=  positionInTH[j1];
					positionInTH[j1] = temp;

					hash[p].posInVocab =j1;
					hash[h].posInVocab =i;

					topBlock[f]=j1+1;

					//	if (hash[h].size<hash[p].size) {
					
					{   unsigned long tmp;
						kkk1=hash[h].codeword;
						kkk2=hash[p].codeword;
						tmp=kkk1;
						if (kkk1 > kkk2) {	
							kkk2=kkk1;
							kkk1=tmp;
						}
						
						if (  ((kkk1 < W[S_value][0]) && (kkk2 >= W[S_value][0])) ||
							  ((kkk1 < W[S_value][1]) && (kkk2 >= W[S_value][1])) ||
							  ((kkk1 < W[S_value][2]) && (kkk2 >= W[S_value][2])) 
						    )				
						kkk1=kkk2-1;
						else kkk1=kkk2;
					}
					
					
					//kkk1= oBytes(S_value,hash[h].codeword);
					//kkk2= oBytes(S_value,hash[p].codeword);
					if (kkk1 < kkk2) {				
						swap(&(hash[p].codeword),&(hash[h].codeword));

						writeBW(bw,(unsigned char *)CC_SWAP,LEN_SCAPES_C_SWAP);	
						#ifdef DEBUG_ENABLED
							SWAPS++;
						#endif
//						getCode(hash[h].codeword);
//						writeCodeWord(bw,&CODEWORD,&K);
						getCodeAndWrite (bw,hash[h].codeword);
					}

					//writes the codeword.
					{					
//						getCode(hash[addrInTH].codeword);
//						writeCodeWord(bw,&CODEWORD,&K);
						getCodeAndWrite (bw,hash[addrInTH].codeword);
					}

				}//end else "j <> zeronode"

			} //End of dynamic update

		/*********************************/
		//getCode(j);
		takeAccount(bw);
		
    }/* inner while*/

	/***************************************************************************/
	/*  main 3.2, first iteration to end the pending buffer  (the first buffer)*/
	/***************************************************************************/


	if (! ONLY1BUFFER){
		nextbuf=pbeg;
		bufLoad (B, (char*)nextbuf);
		//		  fprintf(stderr,"\n READY TO GO MAIN3.2 : GENERAL PROCESS");
	}
	else {
		bufDestroy(B);
		close(f);
		closeBW(bw);

		free(hash);
		free(topBlock); //
		free(positionInTH); // 	
		destroyMemoryManager(_memMgr);
		return;
	}
	
	/***************************************************************************/
	/*  main 3.3, GENERAL COMPRESSION PROCESS, AFTER INITIAL WORK YET DONE     */
	/***************************************************************************/
	
	//fprintf(stderr,"\n MAIN 3 GENERAL");

	if (bufEmpty(B) ) {
		bufDestroy(B);
		close(f);
		closeBW(bw);
	
		free(hash);
		free(topBlock); //
		free(positionInTH); // 
		destroyMemoryManager(_memMgr);
		return;
	}
	//ELSE BYTES REMAINS TO BE READ... AND FIRST IS AN ALREADY READ BYTE
	size = 0;
	
	while (( B->size==B->dsize )  ) {

		bufCurrent (B,(char**)&text,(char**)&top);
		pbeg = text; pend = top;
      	//fprintf(stderr,"\n Entra 1 pbeg [%d] pend [%d] valid [%d]",pbeg,pend, _Valid[*pbeg]);

		while (( pbeg < pend ))  {
			//parsing a word or separator
			size=0;
			wordstart = pbeg;
			if (_Valid[*pbeg]) {   //alphanumerical data
	
				while ( (size<MAX_SIZE_OF_WORD) && (  _Valid[*pbeg] )) {
						size++;
						pbeg++;
				}
	
		    }
	      	else {
				if (*pbeg != ' ') { //a separator comes starting in ' ' comes, so it is a new word
	
					while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg])) {
	
						size++;
						pbeg++;
					}
				}
				else {  //a  SPACE comes, so we have to test if next character is alphanumerical or not
	
					pbeg++;
					if (_Valid [*pbeg] ) {
						wordstart = pbeg;
						while ( (size<MAX_SIZE_OF_WORD) && (_Valid[*pbeg] )) {
							size++;
							pbeg++;
						}
	
					}
					else {   // a "separator word" is parsed until an alphanumerical character is found
	
						size++;
						while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg] )) {
							size++;
							pbeg++;
						}
					}//else {  // a "separator word"
				}// end else  (!bufEmpty(B)) {
	  		}
	

			//The parsed word ...
			aWord=wordstart;
			j = search ((unsigned char *)aWord, size, &addrInTH );
	
			//Dynamic update of the data structures of the compressor//
			{
				register unsigned long p,f,i,h,j1;
				register unsigned long temp;

				p=addrInTH;
		
				if (j==zeroNode) {
					insertElement ((unsigned char *) aWord, size, &addrInTH);
					hash[p].len=size;

					//getCode(j);
					hash[p].codeword = j;
					
					hash[p].weight = 1;
					hash[p].posInVocab = zeroNode;

					positionInTH[zeroNode] = p;
					zeroNode++;

					writeBW(bw,(unsigned char *)CC_NEW,LEN_SCAPES_C_NEW);	
					writeBWPlusZero(bw,(unsigned char *)aWord,size);
				}
				else {
					f = hash[p].weight;
					hash[p].weight +=1;

					i = hash[p].posInVocab;									
					j1 = topBlock[f];
					h = positionInTH[j1];

					temp = positionInTH[i];
					positionInTH[i]=  positionInTH[j1];
					positionInTH[j1] = temp;

					hash[p].posInVocab =j1;
					hash[h].posInVocab =i;

					topBlock[f]=j1+1;

					//	if (hash[h].size<hash[p].size) {
					
					{   unsigned long tmp;
						kkk1=hash[h].codeword;
						kkk2=hash[p].codeword;
						tmp=kkk1;
						if (kkk1 > kkk2) {	
							kkk2=kkk1;
							kkk1=tmp;
						}
						
						if (  ((kkk1 < W[S_value][0]) && (kkk2 >= W[S_value][0])) ||
							  ((kkk1 < W[S_value][1]) && (kkk2 >= W[S_value][1])) ||
							  ((kkk1 < W[S_value][2]) && (kkk2 >= W[S_value][2])) 
						    )				
						kkk1=kkk2-1;
						else kkk1=kkk2;
					}
					
					
					//kkk1= oBytes(S_value,hash[h].codeword);
					//kkk2= oBytes(S_value,hash[p].codeword);
					if (kkk1 < kkk2) {				
						swap(&(hash[p].codeword),&(hash[h].codeword));

						writeBW(bw,(unsigned char *)CC_SWAP,LEN_SCAPES_C_SWAP);	
						#ifdef DEBUG_ENABLED
							SWAPS++;
						#endif
//						getCode(hash[h].codeword);
//						writeCodeWord(bw,&CODEWORD,&K);
						getCodeAndWrite (bw,hash[h].codeword);
					}

					//writes the codeword.
					{					
//						getCode(hash[addrInTH].codeword);
//						writeCodeWord(bw,&CODEWORD,&K);
						getCodeAndWrite (bw,hash[addrInTH].codeword);
					}

				}//end else "j <> zeronode"

			} //End of dynamic update
	
			/*********************************/
			//getCode(j);
			takeAccount(bw);
			
	    }/* inner while*/


		//fprintf(stderr,"\n Procesa la palabra1 [%s]",hash[addrInTH].word);
		nextbuf=pbeg;
		bufLoad (B, (char*)nextbuf);

	}//while( OUTER 3.3)

	//fprintf(stderr,"\n\n Full buffers already processed (zeronode %ld",zeroNode);


	/***************************************************************************/
	/*  main 3.4, To process THE last buffer                                   */
	/***************************************************************************/
	if (!bufEmpty(B)){

	 	bufCurrent (B,(char**)&text,(char**)&top);
	 	pbeg = text; pend = top;

		while (( pbeg < pend ))  {
			//parsing a word or separator
			size=0;
			wordstart = pbeg;
			if (_Valid[*pbeg]) {   //alphanumerical data
	
				while ( (size<MAX_SIZE_OF_WORD) && (  _Valid[*pbeg] ) ) {
						size++;
						pbeg++;
				}
	
		    }
	      	else {
				if (*pbeg != ' ') { //a separator comes starting in ' ' comes, so it is a new word
	
					while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg])) {
	
						size++;
						pbeg++;
					}
				}
				else {  //a  SPACE comes, so we have to test if next character is alphanumerical or not
	
					pbeg++;
					if (_Valid [*pbeg] ) {
						wordstart = pbeg;
						while ( (size<MAX_SIZE_OF_WORD) && (_Valid[*pbeg] )) {
							size++;
							pbeg++;
						}
	
					}
					else {   // a "separator word" is parsed until an alphanumerical character is found
	
						size++;
						while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg] )) {
							size++;
							pbeg++;
						}
					}//else {  // a "separator word"
				}// end else  (!bufEmpty(B)) {
	  		}
	  		
	  		if ( pbeg >= pend ) {  //need in case of the last word been parsed further away of cend. 
	  			size -= (pbeg-pend);	
	  		}
	
			//The parsed word ...
			aWord=wordstart;
			j = search ((unsigned char *)aWord, size, &addrInTH );
	
			//Dynamic update of the data structures of the compressor//
			{
				register unsigned long p,f,i,h,j1;
				register unsigned long temp;

				p=addrInTH;
		
				if (j==zeroNode) {
					insertElement ((unsigned char *) aWord, size, &addrInTH);
					hash[p].len=size;

					//getCode(j);
					hash[p].codeword = j;
					
					hash[p].weight = 1;
					hash[p].posInVocab = zeroNode;

					positionInTH[zeroNode] = p;
					zeroNode++;

					writeBW(bw,(unsigned char *)CC_NEW,LEN_SCAPES_C_NEW);	
					writeBWPlusZero(bw,(unsigned char *)aWord,size);
				}
				else {
					f = hash[p].weight;
					hash[p].weight +=1;

					i = hash[p].posInVocab;									
					j1 = topBlock[f];
					h = positionInTH[j1];

					temp = positionInTH[i];
					positionInTH[i]=  positionInTH[j1];
					positionInTH[j1] = temp;

					hash[p].posInVocab =j1;
					hash[h].posInVocab =i;

					topBlock[f]=j1+1;

					//	if (hash[h].size<hash[p].size) {
					
					{   unsigned long tmp;
						kkk1=hash[h].codeword;
						kkk2=hash[p].codeword;
						tmp=kkk1;
						if (kkk1 > kkk2) {	
							kkk2=kkk1;
							kkk1=tmp;
						}
						
						if (  ((kkk1 < W[S_value][0]) && (kkk2 >= W[S_value][0])) ||
							  ((kkk1 < W[S_value][1]) && (kkk2 >= W[S_value][1])) ||
							  ((kkk1 < W[S_value][2]) && (kkk2 >= W[S_value][2])) 
						    )				
						kkk1=kkk2-1;
						else kkk1=kkk2;
					}
					
					
					//kkk1= oBytes(S_value,hash[h].codeword);
					//kkk2= oBytes(S_value,hash[p].codeword);
					if (kkk1 < kkk2) {				
						swap(&(hash[p].codeword),&(hash[h].codeword));

						writeBW(bw,(unsigned char *)CC_SWAP,LEN_SCAPES_C_SWAP);	
						#ifdef DEBUG_ENABLED
							SWAPS++;
						#endif
//						getCode(hash[h].codeword);
//						writeCodeWord(bw,&CODEWORD,&K);
						getCodeAndWrite (bw,hash[h].codeword);
					}

					//writes the codeword.
					{					
//						getCode(hash[addrInTH].codeword);
//						writeCodeWord(bw,&CODEWORD,&K);
						getCodeAndWrite (bw,hash[addrInTH].codeword);
					}

				}//end else "j <> zeronode"

			} //End of dynamic update
			/*********************************/
			//getCode(j);
			takeAccount(bw);
			
	    }/* inner while*/
	}//if (!bufEmpty(B){



	MAXFREQ=hash[positionInTH[0]].weight;
	bufDestroy(B);
	close(f);
	closeBW(bw);

	free(hash);
	free(topBlock); //
	free(positionInTH); // 	
	destroyMemoryManager(_memMgr);

} //mainCompressor

/*------------------------------------------------------------------
 Obtains the size of the file to compress.
------------------------------------------------------------------ */
unsigned long fileSize (char *filename){
	FILE *fpText;
	unsigned long fsize;
	fpText = fopen(filename,"rb");
	fsize=0;
	if (fpText) {
		fseek(fpText,0,2);
		fsize= ftell(fpText);
		fclose(fpText);
		fprintf(stderr,"fileSize = %ld",fsize);
	}
	return fsize;
}

/*------------------------------------------------------------------
 Obtains the size of the file to compress.
------------------------------------------------------------------ */
void writeSizeVocabulary (char *filename, unsigned long valueSIZEVOC) {
	char fileNametmp[24];
	int file;
	
	strcpy (fileNametmp,filename);
	strcat(fileNametmp,".voc");
	
	if ( (file = open (fileNametmp, O_WRONLY|O_CREAT, 0666)) ==-1)
	{
		fprintf(stderr,"\nError while opening file: %s\n exiting...\n",fileNametmp);
		exit(0);
	}
	else  fprintf(stderr,"\nFILE: %s OPENED OK\n",fileNametmp);

	fprintf(stderr,"   *Saving: VOCSIZE=%ld !\n",valueSIZEVOC);		
	write(file,  &valueSIZEVOC,  4);
	close(file);
}

/*------------------------------------------------------------------
  Shows a message with info of this version of the code 
  ------------------------------------------------------------------*/
void showMessage() {
	printf("\n*Dynamic Lightweight SC Dense Code v0.9 [dynamic compressor]");
	printf("\n*CopyRight(c)2006  Nieves R. Brisaboa, Gonzalo Navarro, Antonio Faria and Jos R. Param\n");
	
    printf("\n  #DLSCDC comes with ABSOLUTELY NO WARRANTY; ");
    printf("\n  #This is free software, and you are welcome to redistribute it ");
    printf("\n  #under certain conditions\n\n");
}

/*------------------------------------------------------------------ */
/*
 *  COMPRESSOR PROGRAM OF DYNAMIC ETDC.
 */
/*------------------------------------------------------------------ */

int main(int argc, char* argv[])
{
	char fileName[20];
	long sizeNValue,sizeNValue2;
	unsigned long bytesFile;

	StartRusageTime();

	showMessage();

	if (argc < 3) {
		fprintf(stderr,"\n Wrong call to DSCTDC\n");
		fprintf(stderr," Use:  %s  <in file>  <out file> [<VOC SIZE> <TOP SIZE>] \n",argv[0]);
		return 0;
	}

	sizeNValue=MAX_LEN_VOCABULARY;
	sizeNValue2=INITIAL_TOP_FREQ;

	//Estimation (Using Heap's law) of the size of the vocabulary (in number of distinct words).
	//sizeNValue=N_value;
	bytesFile = fileSize(argv[1]);
	if (!bytesFile) {
		fprintf(stderr, "\nFILE EMPTY OR FILE NOT FOUND !!\nCompression aborted\n");
		exit(0);
	}

	if(bytesFile<5000000) bytesFile = 5000000;
	sizeNValue = (unsigned long) floor(3.9* pow(bytesFile,0.69) );
	sizeNValue2 = sizeNValue *4;
	
	if (sizeNValue > MAX_LEN_VOCABULARY) {sizeNValue = MAX_LEN_VOCABULARY;}

	if (argc ==5) {
		  sizeNValue = atol(argv[3]);
		  sizeNValue2 = atol(argv[4]);
	}

	fprintf(stderr,"\nMax size of voc set[%ld] and Max Freq [%ld] allowed",sizeNValue,sizeNValue2);

	mainCompressor(argv[1], argv[2],sizeNValue,sizeNValue2);
	
	//uses an external file to store the number of words in the vocabulary.
	writeSizeVocabulary (argv[2], zeroNode);
	
	StopRusageTime("timesCDLSCDC",argv[1]);
	fprintf(stderr,"\nExact Number of words in vocabulary: %ld ",zeroNode);
	fprintf(stderr,"\nFinal maximum word frequency: %ld ",MAXFREQ);
	#ifdef DEBUG_ENABLED
		fprintf(stderr,"\nExact Number of swaps: %ld ",SWAPS);
		fprintf(stderr,"\nExact Number of Schanges: %ld ",SCHANGES);
	#endif

	fprintf(stderr,"\n\n*Compression finished.\n");
}



